open Big_int

module type FIELD = sig
  type t 

  val zero : t
  val one : t

  val add : t -> t -> t

  val multiply : t -> t -> t

  val negate : t -> t

  (* return the multiplicative inverse of the argument.
   * Behavior is undefined if the argument is zero.
   *)
  val inverse : t -> t

  val to_string : t -> string

  val num_of_big_int : big_int -> t

  val num_of_int : int -> t
end

module type OrderedField = sig
  include FIELD
  type cmp = Less | Greater | Equal
  val cmp : t -> t -> cmp
end

module Floats : OrderedField with type t = float = struct
  type t = float
  type cmp = Less | Greater | Equal
  let zero = 0.
  let one = 1.
  let add x= failwith ""
  let multiply x= failwith ""
  let negate x= failwith ""
  let inverse x= failwith ""
  let cmp x= failwith ""
  let to_string x= failwith ""
  let num_of_big_int x= failwith ""
  let num_of_int x= failwith ""
end

module Q : OrderedField with type t = big_int * big_int = 
struct
(* Invariant: denominator is always positive *)
  type t = big_int * big_int
  type cmp = Less | Greater | Equal

  let ( * ) = mult_big_int
  let (+) = add_big_int
  let (-) = sub_big_int
  let (/) = div_big_int

  let zero = zero_big_int, unit_big_int
  let one = unit_big_int, unit_big_int

  let reduce (a,b) = 
    let gcd = gcd_big_int a b in
    (a / gcd, b / gcd)

  let add (a1, a2) (b1, b2) = 
    reduce (a1 * b2 + a2 * b1, a2 * b2)

  let multiply (a1, a2) (b1, b2) = reduce (a1 * b1, a2 * b2)

  let negate (xn, xd) = 
    if le_big_int xd zero_big_int
    then (xn, (big_int_of_int (-1)) * xd)
    else ((big_int_of_int (-1)) * xn, xd)

  let inverse (x1, x2) = 
    let check_pos_numerator = compare_big_int x1 zero_big_int in
    if check_pos_numerator < 0 then
      ((big_int_of_int (-1)) * x2 , (big_int_of_int (-1)) * x1)
    else if check_pos_numerator > 0 then (x2, x1)
    else failwith "Zero does not have an inverse"

  let cmp (an, ad) (bn, bd) =     
    let diff = compare_big_int (an*bd) (ad*bn) in
    if diff = 0 then Equal
    else if diff > 0 then Greater
    else Less

  let to_string (xn, xd) = 
    (string_of_big_int xn) ^ " / " ^ (string_of_big_int xd)

  let num_of_big_int x= (x, big_int_of_int 1)

  let num_of_int x= (big_int_of_int x, big_int_of_int 1)
end

module R : OrderedField with type t = big_int -> Q.t = 
struct
  type t = big_int -> Q.t
  type cmp = Less | Greater | Equal

  let ( * ) = mult_big_int
  let (+) = add_big_int
  let (-) = sub_big_int
  let (/) = div_big_int

  let zero = fun n -> Q.zero

  let one  = fun n -> Q.one
  
  let add x y = 
    fun n ->
      let m = mult_int_big_int 2 n in
      Q.add (x m) (y m)
     
  (* [Normalize x n] = a where a/2n is within 1/n of the rational x(n) *)
  let normalize x n = 
    let (a,b) = x n in
    ((mult_int_big_int 2 n) * a) / b

  let canonical_bound x = 
    let b = abs_big_int (normalize x (big_int_of_int 1)) in
    ((add_int_big_int 4 b) / (big_int_of_int 2))

  let multiply x y = 
    let k = max_big_int (canonical_bound x) (canonical_bound y) in
    fun n -> 
      let m = (mult_int_big_int 2 k) * n in
           Q.multiply (x m) (y m)

  let negate x = fun n -> Q.negate (x n)

  (* inverse of a real x # 0 *)
  (* we first have to find an N such that |x N| > 1/N *)
  let find_least test =
    let rec aux n = if test n then n else aux (succ_big_int n) in
    aux (big_int_of_int 1)

  let inv_bound x = 
    let test n = let (a,b) = x n in
              lt_big_int (abs_big_int b) (abs_big_int (n * a))
    in find_least test

  let inverse x = 
    let b = inv_bound x in
    let b2 = mult_big_int b b in
    let b3 = mult_big_int b b2 in
    fun n ->
      if lt_big_int n b
      then Q.inverse (x b3) 
      else let m = mult_big_int n b2 in
           Q.inverse (x m) 

  let cmp x y= 
    let diff = add x (negate y) in
    (* simultaneously check if diff or negate diff is positive *)
    let rec helper n = 
      let (a,b) = diff n in
      if lt_big_int b (n * a) then Greater
      else if lt_big_int b (mult_int_big_int (-1) (n * a)) then Less
      else helper (succ_big_int n) in
    helper (big_int_of_int 1)

  let to_string x = failwith ""

  let num_of_big_int x = fun n -> Q.num_of_big_int x

  let num_of_int x = fun n -> Q.num_of_int x
end